// Copyright 2019-2021 Graham Clark. All rights reserved.  Use of this source
// code is governed by the MIT license that can be found in the LICENSE
// file.

package system

var CopyToClipboard = []string{"termux-clipboard-set"}

var OpenURL = []string{"am", "start", "-a", "android.intent.action.VIEW", "-d"}
